/* CmdBkg (c) 2016-2020 Mikael Sollenborn and Steffen Ilhardt */

#undef WINVER
#define WINVER 0x0501
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#define _WIN32_DCOM

// Usage: cmdbkg "file.[bmp|jpg|gif|png]" [transparency 0-100(default 33) [includeBorders]]
// Specify no arguments to remove previous background.
// Specify /? as first argument to see this help.

// Compilation: gcc -o cmdbkg.exe cmdbkg.c -lgdi32 -ldwmapi -lgdiplus -s

// TODO/Issues:
// 1. Allow centered placement of image, no stretching?
// 2. Background window visible a short time after restoring minimized console window. Minimize background if console window is minimized?
// 3. Inelegant polling. Use SetWinEventHook instead?

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <dwmapi.h>

// GDI+ headers are C++-only in MS Visual Studio (even dependencies of the flat API).
// Instead of including <gdiplus.h> we declare the necessary things here. However the gdiplus lib has to be linked.
typedef struct tag_GdiplusStartupInput
{
  UINT32 GdiplusVersion;
  void *DebugEventCallback;
  BOOL SuppressBackgroundThread;
  BOOL SuppressExternalCodecs;
} GdiplusStartupInput;

int __stdcall GdiplusStartup(ULONG_PTR *, const GdiplusStartupInput *, void *);
void __stdcall GdiplusShutdown(ULONG_PTR);
int __stdcall GdipCreateBitmapFromFile(const wchar_t *, void **);
int __stdcall GdipCreateHBITMAPFromBitmap(void *, HBITMAP *, DWORD);

#define POLL_INTERVAL 20U
#define TIMEOUT 2000U
#define DEFAULT_TRANSPARENCY 33L
#define GESSO COLOR_3DLIGHT
#define SECOND_RUN L"_secondRun_"

// Read the content of an image file, convert it to a bitmap in memory, and return a handle of it. Store the handle for subsequent calls.
static HBITMAP GetHBitmapFromFile(const wchar_t *imagePath)
{
  static HBITMAP hBmp = NULL;
  if (hBmp || !imagePath)
    return hBmp;

  const GdiplusStartupInput gpStartupInput = { 1U, NULL, FALSE, FALSE };
  ULONG_PTR gpToken = 0U;
  GdiplusStartup(&gpToken, &gpStartupInput, NULL);
  void *bmpPtr = NULL;
  GdipCreateBitmapFromFile(imagePath, &bmpPtr);
  if (bmpPtr)
    GdipCreateHBITMAPFromBitmap(bmpPtr, &hBmp, GetSysColor(GESSO));

  GdiplusShutdown(gpToken);
  return hBmp;
}

// Load an image file and insert it into a window. Parameter imagePath is used only once.
static BOOL LoadImg(HWND hWnd, const wchar_t *imagePath, POINT pos, SIZE size, long zoom)
{
  BOOL ret = FALSE;
  if (hWnd)
  {
    HDC hDc = GetDC(hWnd);
    if (hDc)
    {
      HDC hDcBmp = CreateCompatibleDC(hDc);
      if (hDcBmp)
      {
        HBITMAP hBmp1 = GetHBitmapFromFile(imagePath);
        if (hBmp1)
        {
          BITMAP bmp = { 0 };
          if (GetObjectW(hBmp1, sizeof(bmp), &bmp))
          {
            if (size.cx == -1L)
            {
              if ((size.cx = (long)((double)bmp.bmWidth * zoom / 100.0 + 0.5)) <= 0 || (size.cy = (long)((double)bmp.bmHeight * zoom / 100.0 + 0.5)) <= 0)
              {
                size.cx = bmp.bmWidth;
                size.cy = bmp.bmHeight;
              }
            }
            HBITMAP hBmp2 = (HBITMAP)CopyImage((HANDLE)hBmp1, IMAGE_BITMAP, size.cx, size.cy, LR_DEFAULTCOLOR);
            if (hBmp2)
            {
              HGDIOBJ hGdiObj = SelectObject(hDcBmp, hBmp2);
              if (hGdiObj != NULL && hGdiObj != HGDI_ERROR)
              {
                if (BitBlt(hDc, (int)pos.x, (int)pos.y, (int)size.cx, (int)size.cy, hDcBmp, 0, 0, SRCCOPY))
                  ret = TRUE;

                DeleteObject(hGdiObj);
              }
              DeleteObject(hBmp2);
            }
          }
          // We do not delete bmp1 here. It's static in GetHBitmapFromFile and must remain valid for later use.
        }
        ReleaseDC(hWnd, hDcBmp);
      }
      ReleaseDC(hWnd, hDc);
    }
  }
  return ret;
}

// Make a console a layered window and let it be tranparent to the specified extend.
static BOOL SetConsoleTransparency(HWND hWnd, long percentage)
{
  if (hWnd && percentage > -1L && percentage < 101L)
  {
    LONG_PTR newWndLong = GetWindowLongPtrW(hWnd, GWL_EXSTYLE) | WS_EX_LAYERED;
    if (!SetWindowLongPtrW(hWnd, GWL_EXSTYLE, newWndLong))
      return FALSE;

    return SetLayeredWindowAttributes(hWnd, 0UL, (BYTE)((255L * (100L - percentage) + 50L) / 100L), LWA_ALPHA);
  }

  return FALSE;
}

// Find the window where the user data field contains the console window handle.
static BOOL CALLBACK CloseWindowByUserData(HWND hwnd, LPARAM lParamConWnd)
{
  if (GetWindowLongPtrW(hwnd, GWLP_USERDATA) == lParamConWnd)
  {
    SendNotifyMessageW(hwnd, WM_CLOSE, 0U, 0);
    for (unsigned i = 0U; i++ < TIMEOUT / POLL_INTERVAL && (IsWindow(hwnd) || IsIconic(hwnd)); Sleep(POLL_INTERVAL)) // Ensure the window has been closed before we return.
      ;

    return FALSE;
  }

  return TRUE;
}

// Window callback function necessary to handle WM_CLOSE and WM_DESTROY.
static LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  switch (msg)
  {
  case WM_CLOSE:
    DestroyWindow(hwnd);
    return 0;
  case WM_DESTROY:
    PostQuitMessage(0);
    return 0;
  default:
    break;
  }

  return DefWindowProcW(hwnd, msg, wparam, lparam);
}

int main(void)
{
  const HWND consoleWnd = GetConsoleWindow();
  if (!consoleWnd || !SendMessageW(consoleWnd, WM_GETICON, 0U, 0)) // The latter roughly verifies that the window is a console (rather than a terminal emulator).
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to prepare window creation (1).", L"Error", MB_ICONERROR);
    return 1;
  }

  int argCount = 0;
  const wchar_t *oldCmdLine = GetCommandLineW();
  wchar_t **argList = CommandLineToArgvW(oldCmdLine, &argCount);
  if (argList == NULL)
  {
    fputs("Cmdbkg error: Unable to get arguments.\n", stderr);
    return 1;
  }

  if (argCount == 2 && wcscmp(argList[1], L"/?") == 0)
  {
    puts("\nCmdBkg v1.2 : Mikael Sollenborn and Steffen Illhardt 2016-2020\n\nUsage: cmdbkg \"file.[bmp|jpg|gif|png]\" [transparency 0-100(default 33) [includeBorders]]\n\nSpecify no arguments to remove previous background.\n");
    LocalFree(argList);
    return 0;
  }

  if (argCount > 1 && wcscmp(argList[1], SECOND_RUN) != 0)
  {
    EnumWindows(CloseWindowByUserData, (LPARAM)consoleWnd); // in case there is already a cmdbkg process running for this console
    const size_t oldLen = wcslen(oldCmdLine);
    const wchar_t scnd[] = L" " SECOND_RUN;
    wchar_t *newCmdLine = malloc(oldLen * sizeof(wchar_t) + sizeof(scnd));
    if (!newCmdLine)
    {
      fputs("Cmdbkg error: Unable to allocate memory.\n", stderr);
      LocalFree(argList);
      return 1;
    }

    // insert two additional argument tokens into the command line, 1st the string "_secondRun_", and 2nd the console window handle as HEX string
    const wchar_t *oldCursor = wcsstr(oldCmdLine, argList[0]) + wcslen(argList[0]);
    LocalFree(argList);
    const wchar_t *tmpPtr = wcspbrk(oldCursor, L" \t");
    oldCursor = tmpPtr ? tmpPtr : oldCmdLine + oldLen;
    wchar_t *newCursor = (wchar_t *)memcpy(newCmdLine, oldCmdLine, (size_t)(oldCursor - oldCmdLine) * sizeof(wchar_t)) + (oldCursor - oldCmdLine);
    newCursor = (wchar_t *)memcpy(newCursor, scnd, sizeof(scnd) - sizeof(wchar_t)) + (sizeof(scnd) - sizeof(wchar_t)) / sizeof(wchar_t);
    memcpy(newCursor, oldCursor, (oldLen - (oldCursor - oldCmdLine) + 1) * sizeof(wchar_t));

    // Start the child process.
    PROCESS_INFORMATION pi = { 0 };
    STARTUPINFOW si = { 0 };
    si.cb = sizeof(si);
    if (!CreateProcessW(NULL, newCmdLine, NULL, NULL, FALSE, 0UL, NULL, NULL, &si, &pi))
    {
      fprintf(stderr, "Cmdbkg error: CreateProcess failed (%lu).\n", GetLastError());
      free(newCmdLine);
      return 1;
    }

    WaitForInputIdle(pi.hProcess, TIMEOUT); // Ensure the child process' window can already be found before we return to our parent.
    Sleep(POLL_INTERVAL); // Extends the previous.
    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
    free(newCmdLine);
    return 0;
  }

  FreeConsole();
  SetConsoleTransparency(consoleWnd, 0L);
  if (argCount < 3)
  {
    EnumWindows(CloseWindowByUserData, (LPARAM)consoleWnd);
    LocalFree(argList);
    return 0;
  }

  WNDCLASSW wndCls = { 0 };
  wndCls.lpszClassName = L"CmdBkgWindowBackground";
  wndCls.lpfnWndProc = WindowProc;
  wndCls.hInstance = NULL;
  wndCls.hbrBackground = (HBRUSH)GESSO;
  wndCls.hCursor = LoadCursor(NULL, IDC_ARROW);
  if (!RegisterClassW(&wndCls))
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to prepare window creation (2).", L"Error", MB_ICONERROR);
    LocalFree(argList);
    return 1;
  }

  POINT wndOrigin = { 0 };
  SIZE wndSize = { 0 };
  RECT wndBounds = { 0 }, extendBounds = { 0 };
  GetWindowRect(consoleWnd, &wndBounds);
  wndOrigin.x = wndBounds.left;
  wndOrigin.y = wndBounds.top;
  wndSize.cx = wndBounds.right - wndBounds.left;
  wndSize.cy = wndBounds.bottom - wndBounds.top;
  DwmGetWindowAttribute(consoleWnd, DWMWA_EXTENDED_FRAME_BOUNDS, &extendBounds, sizeof(extendBounds));
  long frameWidth = 0L, frameHight = 0L, frameTitleHight = 0L;
  if (argCount <= 4 || (argCount > 3 && extendBounds.left != wndBounds.left))
  {
    frameWidth = GetSystemMetrics(SM_CXSIZEFRAME);
    if (argCount <= 4)
      frameTitleHight = GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION);

    frameHight = frameTitleHight + GetSystemMetrics(SM_CYSIZEFRAME);
  }

  const HWND bkgWnd = CreateWindowExW(WS_EX_TOOLWINDOW, wndCls.lpszClassName, NULL, WS_POPUP | WS_VISIBLE, wndOrigin.x + frameWidth, wndOrigin.y + frameTitleHight, wndSize.cx - frameWidth * 2, wndSize.cy - frameHight, NULL, NULL, NULL, NULL);
  if (!bkgWnd)
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to create background window.", L"Error", MB_ICONERROR);
    LocalFree(argList);
    return 1;
  }

  // The user data field contains the console window handle for the sake of finding it to close the background process.
  SetWindowLongPtrW(bkgWnd, GWLP_USERDATA, (LONG_PTR)consoleWnd);
  wndSize.cx -= frameWidth * 2L;
  wndSize.cy -= frameTitleHight;
  if (!LoadImg(bkgWnd, argList[2], (POINT){ 0 }, wndSize, 100L))
  {
    MessageBoxW(NULL, L"Cmdbkg error: Unable to load image.", L"Error", MB_ICONERROR);
    LocalFree(argList);
    return 1;
  }

  ShowWindow(bkgWnd, SW_SHOWNORMAL);
  SetConsoleTransparency(consoleWnd, argCount > 3 ? wcstol(argList[3], NULL, 10) : DEFAULT_TRANSPARENCY);
  LocalFree(argList);
  SetForegroundWindow(consoleWnd);
  HWND currFgWnd = GetForegroundWindow();
  int initialRedraw = 10;
  while (IsWindow(consoleWnd) || IsIconic(consoleWnd))
  {
    MSG msg = { 0 };
    const UINT_PTR timerId = SetTimer(NULL, 0U, POLL_INTERVAL, NULL);
    const BOOL res = GetMessageW(&msg, NULL, 0U, 0U);
    KillTimer(NULL, timerId);
    if (res == 0 || res == -1)
      break;

    DispatchMessageW(&msg);

    const HWND tempWnd = GetForegroundWindow();
    if (currFgWnd != tempWnd)
    {
      currFgWnd = tempWnd;
      if (initialRedraw < 5)
        initialRedraw = 5;
    }

    RECT newBounds = { 0 };
    GetWindowRect(consoleWnd, &newBounds);
    if (initialRedraw > 0 || memcmp(&newBounds, &wndBounds, sizeof(RECT)))
    {
      wndBounds = newBounds;
      wndOrigin.x = wndBounds.left;
      wndOrigin.y = wndBounds.top;
      wndSize.cx = wndBounds.right - wndBounds.left;
      wndSize.cy = wndBounds.bottom - wndBounds.top;
      const LONG_PTR conStyle = GetWindowLongPtrW(consoleWnd, GWL_STYLE);
      if (conStyle & WS_POPUP)
        SetWindowPos(bkgWnd, consoleWnd, wndOrigin.x, wndOrigin.y, wndSize.cx, wndSize.cy, SWP_SHOWWINDOW);
      else
      {
        SetWindowPos(bkgWnd, consoleWnd, wndOrigin.x + frameWidth, wndOrigin.y + frameTitleHight, wndSize.cx - frameWidth * 2, wndSize.cy - frameHight, SWP_SHOWWINDOW);
        wndSize.cx -= frameWidth * 2L;
        wndSize.cy -= frameTitleHight;
      }

      LoadImg(bkgWnd, NULL, (POINT){ 0 }, wndSize, 100L);
      if (initialRedraw > 0)
        --initialRedraw;
    }
  }

  DeleteObject(GetHBitmapFromFile(NULL));
  return 0;
}
